from bluecon.model.AccessDoor import AccessDoor


class Pairing:
    def __init__(self, pairingResponseDictionary: dict):
        self.id = pairingResponseDictionary['id']
        self.deviceId = pairingResponseDictionary['deviceId']
        self.accessDoorMap = {k: AccessDoor(v) for k, v in pairingResponseDictionary['accessDoorMap'].items() if v['visible']}
        pass



from bluecon.BlueConAPI import BlueConAPI
from bluecon.storage.IOAuthTokenStorage import IOAuthTokenStorage
from bluecon.storage.InMemoryOAuthTokenStorage import InMemoryOAuthTokenStorage
from bluecon.storage.INotificationInfoStorage import INotificationInfoStorage
from bluecon.notifications.INotification import INotification
from bluecon.notifications.NotificationBuilder import NotificationBuilder
from bluecon.notifications.CallNotification import CallNotification
from bluecon.notifications.CallEndNotification import CallEndNotification
from bluecon.oauth.OAuthToken import OAuthToken


from bluecon.notifications.CallEndNotification import CallEndNotification
from bluecon.notifications.CallNotification import CallNotification
from bluecon.notifications.INotification import INotification


class NotificationBuilder:
    @classmethod
    def fromNotification(cls, notification: dict, notificationId: str) -> INotification:
        if notification['FermaxNotificationType'] == "Call":
            return CallNotification(notification, notificationId)
        elif notification['FermaxNotificationType'] == "CallEnd":
            return CallEndNotification(notification)
        else:
            raise NotImplementedError


from json import dumps, loads, JSONEncoder
import datetime
from dateutil.parser import isoparse

class OAuthToken:
    def __init__(self, oauth_response: dict):
        self.accessToken = oauth_response['access_token']
        self.refreshToken = oauth_response['refresh_token']
        if (oauth_response.get('expires_in') is not None):
            self.expiresIn = oauth_response['expires_in']
        if (oauth_response.get('expires_on') is not None):
            self.expiresOn = isoparse(oauth_response['expires_on'])
        else:
            self.expiresOn = datetime.datetime.now() + datetime.timedelta(seconds = self.expiresIn)
    
    def getBearerAuthHeader(self) -> dict:
        """Get a dictionary containing the Authorization HTTP header for this token"""

        return {
            'Authorization': f'Bearer {self.accessToken}'
        }
    
    def isExpired(self) -> bool:
        """Check if the token has already expired (for safety, it is considered expired 1 hour before actual expiration)"""

        return datetime.datetime.now() >= (self.expiresOn - datetime.timedelta(hours = 1))
    
    def getRefreshToken(self) -> str:
        """Get the refresh token for this OAuthToken"""
        
        return self.refreshToken

    def toJson(self) -> str:
        return dumps(self, cls = OAuthTokenJSONEncoder)
    
    @classmethod
    def fromJson(csl, json: str):
        return OAuthToken(loads(json))
    
class OAuthTokenJSONEncoder(JSONEncoder):
    def default(self, o: OAuthToken):
        return {
            "access_token": o.accessToken,
            "refresh_token": o.refreshToken,
            "expires_on": o.expiresOn.isoformat()
        }



import abc

from bluecon.oauth.OAuthToken import OAuthToken

class IOAuthTokenStorage(metaclass = abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, __subclass: type) -> bool:
        return hasattr(__subclass, 'retrieveOAuthToken') and callable(__subclass.retrieveOAuthToken) and hasattr(__subclass, 'storeOAuthToken') and callable(__subclass.storeOAuthToken)

    @abc.abstractmethod
    async def retrieveOAuthToken(self) -> OAuthToken | None:
        """Retrieves the OAuthToken from storage"""

        raise NotImplementedError
    
    @abc.abstractmethod
    async def storeOAuthToken(self, oAuthToken: OAuthToken):
        """Stores the OAuthToken"""
        
        raise NotImplementedError


from typing import Any
from bluecon.storage.INotificationInfoStorage import INotificationInfoStorage
import json

class FileNotificationInfoStorage(INotificationInfoStorage):

    async def retrieveCredentials(self) -> dict[str, dict[str, Any]] | None:
        try:
            with open("credentials.json", "r") as f:
                return json.load(f)
        except FileNotFoundError:
            return None
    
    async def storeCredentials(self, credentials: dict[str, dict[str, Any]]):
        with open("credentials.json", "w") as f:
            json.dump(credentials, f)
    
    async def retrievePersistentIds(self) -> list[str] | None:
        try:
            with open("persistent_ids.txt", "r") as f:
                return [x.strip() for x in f]
        except FileNotFoundError:
            return None
    
    async def storePersistentId(self, persistentId: str):
        with open("persistent_ids.txt", "a") as f:
            f.write(persistentId + "\n")


from bluecon.oauth.OAuthToken import OAuthToken
from bluecon.storage.IOAuthTokenStorage import IOAuthTokenStorage


@IOAuthTokenStorage.register
class InMemoryOAuthTokenStorage(IOAuthTokenStorage):
    async def retrieveOAuthToken(self) -> OAuthToken | None:
        if self.__oAuthToken is None:
            return None
        else:
            return OAuthToken.fromJson(self.__oAuthToken)
    
    async def storeOAuthToken(self, oAuthToken: OAuthToken):
        self.__oAuthToken = oAuthToken.toJson()


from datetime import datetime as DateTime
from dateutil.parser import isoparse

class CallLog:
    def __init__(self, callLogResponse: dict):
        self.deviceId: str = callLogResponse['deviceId']
        self.callDate: str = callLogResponse['callDate']
        self.photoId: str | None = callLogResponse['photoId']
    
    def getCallDate(self) -> DateTime:
        return isoparse(self.callDate)



from bluecon.notifications.INotification import INotification


@INotification.register
class CallEndNotification(INotification):
    def __init__(self, notificationData: dict):
        self.deviceId = notificationData['DeviceId']
        self.notificationTitle = notificationData['NotificationTitle']
        self.notificationBody = notificationData['NotificationBody']
        self.callAs = notificationData['CallAs']
    
    def getNotificationType(self) -> str:
        return "CallEnd"
    
    def shouldAcknowledge(self) -> str:
        return False
    
    def getFcmMessageId(self) -> str:
        raise NotImplementedError



from bluecon.notifications.INotification import INotification


@INotification.register
class CallNotification(INotification):
    def __init__(self, notificationData: dict, fcmMessageId: str):
        self.deviceId = notificationData['DeviceId']
        self.fermaxToken = notificationData['FermaxToken']
        self.socketUrl = notificationData['SocketUrl']
        self.notificationTitle = notificationData['NotificationTitle']
        self.notificationBody = notificationData['NotificationBody']
        self.callAs = notificationData['CallAs']
        self.roomId = notificationData['RoomId']
        self.sendAcknowledge = bool(notificationData['SendAcknowledge'])
        self.accessDoorKey = notificationData['AccessDoorKey']
        self.fcmMessageId = fcmMessageId
    
    def getNotificationType(self) -> str:
        return "Call"
    
    def shouldAcknowledge(self) -> str:
        return self.sendAcknowledge
    
    def getFcmMessageId(self) -> str:
        return self.fcmMessageId


class AccessDoor:
    def __init__(self, accessDoorMap: dict):
        self.title = accessDoorMap['title']
        self.block = accessDoorMap['accessId']['block']
        self.subBlock = accessDoorMap['accessId']['subblock']
        self.number = accessDoorMap['accessId']['number']

    def getDirectOpenDoorParamsRequest(self):
        return {
            'block': self.block,
            'number': self.number,
            'subblock': self.subBlock
        }



import asyncio
import aiohttp
import base64
import json
import hashlib
from typing import Callable, List
from threading import Thread

from bluecon.model.AccessDoor import AccessDoor
from bluecon.model.Pairing import Pairing
from bluecon.model.User import User
from bluecon.model.CallLog import CallLog
from bluecon.model.DeviceInfo import DeviceInfo
from bluecon.notifications.INotification import INotification
from bluecon.notifications.NotificationBuilder import NotificationBuilder
from bluecon.oauth.OAuthService import OAuthService
from bluecon.oauth.OAuthToken import OAuthToken
from bluecon.storage.IOAuthTokenStorage import IOAuthTokenStorage
from bluecon.storage.InMemoryOAuthTokenStorage import InMemoryOAuthTokenStorage
from bluecon.storage.INotificationInfoStorage import INotificationInfoStorage
from bluecon.storage.FileNotificationInfoStorage import FileNotificationInfoStorage

from push_receiver import PushReceiver
from push_receiver.android_fcm_register import AndroidFCM

FERMAX_BASE_URL = "https://blue.fermax.io"

class BlueConAPI:
    @classmethod
    async def create(
        cls, 
        username: str, 
        password: str,
        clientId: str,
        clientSecret: str,
        senderId: int,
        apiKey: str,
        projectId: str,
        appId: str,
        packageName: str,
        notificationCallback: Callable[[INotification], None],
        oAuthTokenStorage: IOAuthTokenStorage = InMemoryOAuthTokenStorage(),
        notificationInfoStorage: INotificationInfoStorage = FileNotificationInfoStorage()
    ):
        """Create instance of BlueConAPI for the provided username and password"""

        self = BlueConAPI(clientId, clientSecret, senderId, apiKey, projectId, appId, packageName, notificationCallback, oAuthTokenStorage, notificationInfoStorage)
        oauthToken = await OAuthService.createOAuthToken(self.__getAuthHeader(), username, password)
        await self.__oAuthTokenStorage.storeOAuthToken(oauthToken)
        return self
    
    @classmethod
    async def create_already_authed(
        cls,
        clientId: str,
        clientSecret: str,
        senderId: int,
        apiKey: str,
        projectId: str,
        appId: str,
        packageName: str,
        notificationCallback: Callable[[INotification], None],
        oAuthTokenStorage: IOAuthTokenStorage,
        notificationInfoStorage: INotificationInfoStorage = FileNotificationInfoStorage()
    ):
        """Create instance of BlueConAPI with the OAuth token stored in the provided storage"""
        if await oAuthTokenStorage.retrieveOAuthToken() is not None:
            return BlueConAPI(clientId, clientSecret, senderId, apiKey, projectId, appId, packageName, notificationCallback, oAuthTokenStorage, notificationInfoStorage)
        else:
            raise RuntimeError("Provided IOAuthTokenStorage does not contain a token")
    
    def __init__(
            self, 
            clientId: str,
            clientSecret: str,
            senderId: int,
            apiKey: str,
            projectId: str,
            appId: str,
            packageName: str,
            notificationCallback: Callable[[INotification], None], 
            oAuthTokenStorage: IOAuthTokenStorage,
            notificationInfoStorage: INotificationInfoStorage):
        self.__clientId = clientId
        self.__clientSecret = clientSecret
        self.__senderId = senderId
        self.__apiKey = apiKey
        self.__projectId = projectId
        self.__appId = appId
        self.__packageName = packageName
        self.__oAuthTokenStorage = oAuthTokenStorage
        self.__notificationInfoStorage = notificationInfoStorage
        self.receiver : PushReceiver = None
        self.deviceId : str = None
        self.notificationCallback = notificationCallback
    
    def __getAuthHeader(self) -> str:
        return base64.b64encode(bytes(f'{self.__clientId}:{self.__clientSecret}', 'utf-8')).decode('utf-8')
    
    async def __getOrRefreshOAuthToken(self) -> OAuthToken:
        oAuthToken = await self.__oAuthTokenStorage.retrieveOAuthToken()
        if (oAuthToken.isExpired()):
            oAuthToken = await OAuthService.updateOAuthToken(self.__getAuthHeader(), oAuthToken)
            await self.__oAuthTokenStorage.storeOAuthToken(oAuthToken)
        return oAuthToken
    
    async def getPairings(self) -> List[Pairing]:
        """Get list of pairings for the user"""

        async with aiohttp.ClientSession() as session:
            async with session.get(f'{FERMAX_BASE_URL}/pairing/api/v3/pairings/me', headers = (await self.__getOrRefreshOAuthToken()).getBearerAuthHeader()) as response:
                return list(map(Pairing, json.loads(await response.text())))
    
    async def getUserInfo(self) -> User:
        """Get information about the user"""
        
        async with aiohttp.ClientSession() as session:
            async with session.get(f'{FERMAX_BASE_URL}/user/api/v1/users/me', headers = (await self.__getOrRefreshOAuthToken()).getBearerAuthHeader()) as response:
                return User(json.loads(await response.text()))
    
    async def openDoor(self, deviceId: str, door: AccessDoor) -> bool:
        """Open the provided door"""

        async with aiohttp.ClientSession() as session:
            async with session.post(f'{FERMAX_BASE_URL}/deviceaction/api/v1/device/{deviceId}/directed-opendoor',
                                    json = door.getDirectOpenDoorParamsRequest(),
                                    headers = (await self.__getOrRefreshOAuthToken()).getBearerAuthHeader()) as response:
                return response.status == 200
    
    async def acknowledgeNotification(self, notification: INotification) -> bool:
        """Acknowledges the provided notification if it should be acknowledged"""

        if notification.shouldAcknowledge():
            async with aiohttp.ClientSession() as session:
                async with session.post(f'{FERMAX_BASE_URL}/callmanager/api/v1/message/ack',
                                        json = {
                                            "attended": True,
                                            "fcmMessageId": notification.getFcmMessageId()
                                        },
                                        headers = (await self.__getOrRefreshOAuthToken()).getBearerAuthHeader()) as response:
                    return response.status == 200

    async def registerAppToken(self, active: bool) -> bool:
        async with aiohttp.ClientSession() as session:
            async with session.post(f'{FERMAX_BASE_URL}/notification/api/v1/apptoken',
                                    json = {
                                        "token": self.deviceId,
                                        "appVersion": "3.3.2",
                                        "locale": "en",
                                        "os": "Android",
                                        "osVersion": "Android 13",
                                        "active": active
                                    },
                                    headers = (await self.__getOrRefreshOAuthToken()).getBearerAuthHeader()) as response:
                return response.status == 200
    
    def startNotificationListener(self):
        """Starts the notification listener to get notifications about calls"""

        def listener_thread(blueConAPIClient: BlueConAPI):

            def buildPackageCert():
                sha = hashlib.sha512()
                sha.update(str(blueConAPIClient.__senderId).encode('utf-8'))
                sha.update(blueConAPIClient.__appId.encode('utf-8'))
                sha.update(blueConAPIClient.__apiKey.encode('utf-8'))
                sha.update(blueConAPIClient.__projectId.encode('utf-8'))
                sha.update(blueConAPIClient.__packageName.encode('utf-8'))
                return sha.hexdigest()

            PACKAGE_CERT = buildPackageCert()

            credentials = asyncio.run(blueConAPIClient.__notificationInfoStorage.retrieveCredentials())
            if credentials is None:
                credentials = AndroidFCM.register(
                    api_key=blueConAPIClient.__apiKey,
                    project_id=blueConAPIClient.__projectId,
                    gcm_sender_id = blueConAPIClient.__senderId, 
                    gms_app_id = blueConAPIClient.__appId,
                    android_package_name=blueConAPIClient.__packageName,
                    android_package_cert=PACKAGE_CERT
                )
                asyncio.run(blueConAPIClient.__notificationInfoStorage.storeCredentials(credentials))
            
            blueConAPIClient.deviceId = credentials["fcm"]["token"]
            asyncio.run(blueConAPIClient.registerAppToken(True))
            
            received_persistent_ids = asyncio.run(blueConAPIClient.__notificationInfoStorage.retrievePersistentIds())

            if received_persistent_ids is None:
                blueConAPIClient.receiver = PushReceiver(credentials)
            else:
                blueConAPIClient.receiver = PushReceiver(credentials, received_persistent_ids)

            blueConAPIClient.receiver.listen(on_notification, blueConAPIClient)
        
        def on_notification(blueConAPIClient: BlueConAPI, notification: dict, data_message):
            idstr = data_message.persistent_id

            received_persistent_ids = asyncio.run(blueConAPIClient.__notificationInfoStorage.retrievePersistentIds())

            if received_persistent_ids is not None and any(idstr in x for x in received_persistent_ids):
                return
            
            asyncio.run(blueConAPIClient.__notificationInfoStorage.storePersistentId(idstr))
            
            blueConNotification = NotificationBuilder.fromNotification(notification, data_message.id)
            asyncio.run(blueConAPIClient.acknowledgeNotification(blueConNotification))
            blueConAPIClient.notificationCallback(blueConNotification)
        
        self.__listenerThread = Thread(target = listener_thread, args = (self, ))
        self.__listenerThread.daemon = True
        self.__listenerThread.start()
    
    async def stopNotificationListener(self) -> bool:
        self.__listenerThread.join(10.0)
        await self.registerAppToken(False)
        return self.__listenerThread.is_alive()
    
    async def getLastPicture(self, deviceId: str) -> bytes | None:
        async with aiohttp.ClientSession() as session:
            async with session.get(f'{FERMAX_BASE_URL}/callManager/api/v1/callregistry/participant',
                                    params = {
                                        "appToken": self.deviceId,
                                        "callRegistryType": "all"
                                    },
                                    headers = (await self.__getOrRefreshOAuthToken()).getBearerAuthHeader()) as response:
                responseJson = await response.json()
            callLogs: List[CallLog] = [callLog for callLog in map(CallLog, responseJson) if callLog.deviceId == deviceId and callLog.photoId is not None]
            latestCallLog : CallLog | None = max(callLogs or None, key = lambda x: x.getCallDate())

            if latestCallLog is not None:
                async with session.get(f'{FERMAX_BASE_URL}/callManager/api/v1/photocall',
                                       params = {
                                           "photoId": latestCallLog.photoId
                                       },
                                       headers = (await self.__getOrRefreshOAuthToken()).getBearerAuthHeader()) as response:
                    return base64.b64decode((await response.json())["image"]["data"])
            else:
                return None
    
    async def getDeviceInfo(self, deviceId: str) -> DeviceInfo | None:
        async with aiohttp.ClientSession() as session:
            async with session.get(f'{FERMAX_BASE_URL}/deviceaction/api/v1/device/{deviceId}',
                                   headers = (await self.__getOrRefreshOAuthToken()).getBearerAuthHeader()) as response:
                if response.status == 200:
                    return DeviceInfo(await response.json())
                else:
                    return None



import abc
from typing import Any

class INotificationInfoStorage(metaclass = abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, __subclass: type) -> bool:
        return (
            hasattr(__subclass, 'retrieveCredentials') and callable(__subclass.retrieveCredentials) 
            and hasattr(__subclass, 'storeCredentials') and callable(__subclass.storeCredentials)
            and hasattr(__subclass, 'retrievePersistentIds') and callable(__subclass.retrievePersistentIds) 
            and hasattr(__subclass, 'storePersistentId') and callable(__subclass.storePersistentId)
        )

    @abc.abstractmethod
    async def retrieveCredentials(self) -> dict[str, dict[str, Any]] | None:
        """Retrieves the credentials from storage"""

        raise NotImplementedError
    
    @abc.abstractmethod
    async def storeCredentials(self, credentials: dict[str, dict[str, Any]]):
        """Stores the credentials"""
        
        raise NotImplementedError
    
    @abc.abstractmethod
    async def retrievePersistentIds(self) -> list[str] | None:
        """Retrieves the persistent IDs from storage"""

        raise NotImplementedError
    
    @abc.abstractmethod
    async def storePersistentId(self, persistentId: str):
        """Stores the persistent ID"""
        
        raise NotImplementedError


class DeviceInfo:
    def __init__(self, deviceInfoResponse: dict):
        self.deviceId = deviceInfoResponse['deviceId']
        self.connectionState = deviceInfoResponse['connectionState']
        self.family = deviceInfoResponse['family']
        self.type = deviceInfoResponse['type']
        self.subType = deviceInfoResponse['subtype']
        self.photoCaller = deviceInfoResponse['photocaller']
        self.wirelessSignal = deviceInfoResponse['wirelessSignal']


from bluecon.oauth.OAuthToken import OAuthToken
from aiohttp import ClientSession, FormData

FERMAX_OAUTH_BASE_URL = "https://oauth-blue.fermax.io"

class OAuthService:
    @classmethod
    async def createOAuthToken(cls, authHeader: str, username: str, password: str) -> OAuthToken:
        """Call Fermax Blue OAuth endpoint and generate an OAuthToken from the result"""

        async with ClientSession() as session:
            async with session.post(f'{FERMAX_OAUTH_BASE_URL}/oauth/token',
                                data = FormData(fields = {
                                    "username": username,
                                    "password": password,
                                    "grant_type": "password"
                                }),
                                headers = {
                                    "Authorization": f'Basic {authHeader}'
                                }) as response:
                if response.status == 200:
                    return OAuthToken.fromJson(await response.text())
                else:
                    raise RuntimeError(f"Could not login: {(await response.text())}")
    
    @classmethod
    async def updateOAuthToken(cls, authHeader: str, oAuthToken: OAuthToken) -> OAuthToken:
        """Refresh the OAuthToken by calling the Fermax Blue OAuth endpoint and generating a new OAuthToken from the result"""
        
        async with ClientSession() as session:
            async with session.post(f'{FERMAX_OAUTH_BASE_URL}/oauth/token',
                                    data = FormData(fields = {
                                        "refresh_token": oAuthToken.getRefreshToken(),
                                        "grant_type": "refresh_token"
                                    }),
                                    headers = {
                                        "Authorization": f'Basic {authHeader}'
                                    }) as response:
                return OAuthToken.fromJson(await response.text())


from json import JSONEncoder

from bluecon.model.AccessDoor import AccessDoor
from bluecon.model.Pairing import Pairing
from bluecon.model.User import User
from bluecon.model.DeviceInfo import DeviceInfo

class BlueConJSONEncoder(JSONEncoder):
    def default(self, o):
        if type(o) is User:
            return {
                "email": o.email,
                "locale": o.locale
            }
        elif type(o) is Pairing:
            return {
                "id": o.id,
                "deviceId": o.deviceId,
                "accessDoorMap":o.accessDoorMap
            }
        elif type(o) is AccessDoor:
            return {
                'title': o.title,
                'accessId': {
                    'block': o.block,
                    'subblock': o.subBlock,
                    'number': o.number
                },
                'visible': o.visible
            }
        elif type(o) is DeviceInfo:
            return {
                'deviceId': o.deviceId,
                'connectionState': o.connectionState,
                'family': o.family,
                'type': o.type,
                'subType': o.subType,
                'photoCaller': o.photoCaller,
                'wirelessSignal': o.wirelessSignal
            }
        else:
            raise TypeError


import abc

class INotification(metaclass = abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, __subclass: type) -> bool:
        return hasattr(__subclass, 'getNotificationType') and callable(__subclass.getNotificationType)

    @abc.abstractmethod
    def getNotificationType(self) -> str:
        """Retrieves the notification type for this notification"""

        raise NotImplementedError

    @abc.abstractmethod
    def shouldAcknowledge(self) -> str:
        """Should this notification be acknowledged"""

        raise NotImplementedError
    
    @abc.abstractmethod
    def getFcmMessageId(self) -> str:
        """Retrieves the FCM message ID for this notification"""

        raise NotImplementedError
